import numpy as np
import ctypes
import sys


class CtypeField():
  def __init__(self, field_name, dtype, shape=None, ndim=None):
    self.field_name = field_name
    self.dtype = dtype
    self.shape = shape
    self.ndim = ndim
    if shape != None:
      self.ndim = len(shape)

  def to_input_tuple(self):    
    if self.ndim == None:
      data = (self.field_name, self.dtype)
    else:
      if self.ndim == 1:
        pointer = ctypes.POINTER(self.dtype)
      elif self.ndim == 2:
        pointer = ctypes.POINTER(self.dtype * self.shape[0])
      elif self.ndim == 3:
        pointer = ctypes.POINTER(self.dtype * self.shape[0] * self.shape[1])    
      else:
        raise Exception()
      data = (self.field_name, pointer)
    return data

  def to_output_tuple(self):    
    if self.ndim == None:
      data = (self.field_name, self.dtype)
    else:
      np_dtype = to_numpy_dtype(self.dtype)
      pointer = np.ctypeslib.ndpointer(dtype=np_dtype, ndim=self.ndim)
      data = (self.field_name, pointer)
    return data

class StructFactory():

  def _create_struct(self, ctype_fields, struct_fields):
    class Struct( ctypes.Structure ):
      _fields_ = struct_fields

      def __init__(self):
        self.fields = {}     
        for ctype_field in ctype_fields:
          self.fields[ctype_field.field_name] = ctype_field   
    return Struct() 
  
  def create_input_struct(self, ctype_fields):
    input_struct = self._create_struct(ctype_fields, [ctype_field.to_input_tuple() for ctype_field in ctype_fields])
    for ctype_field in ctype_fields:
      input_struct.fields[ctype_field.field_name] = ctype_field
      if ctype_field.shape != None:
        if ctype_field.ndim == 1:
          setattr(input_struct, ctype_field.field_name, (ctype_field.dtype * ctype_field.shape[0])())
        elif ctype_field.ndim == 2:
          setattr(input_struct, ctype_field.field_name, ( (ctype_field.dtype * ctype_field.shape[0]) * ctype_field.shape[1] )())
        elif ctype_field.ndim == 3:
          setattr(input_struct, ctype_field.field_name, ( ( ( (ctype_field.dtype * ctype_field.shape[0]) * ctype_field.shape[1] ) * ctype_field.shape[2] ) )())
 
    return input_struct


  def create_output_struct(self, ctype_fields):
    return self._create_struct(ctype_fields, [ctype_field.to_output_tuple() for ctype_field in ctype_fields])


def to_numpy_dtype(dtype):
  if dtype == ctypes.c_int:
    return np.int32
  elif dtype == ctypes.c_double:
    return np.double              

def update_struct(src_dict, struct):
  for field_name, field_type in struct._fields_:
    if field_name in src_dict:
      np_dtype = to_numpy_dtype(struct.fields[field_name].dtype)  
      if field_type in [ctypes.c_double, ctypes.c_int]:
        value = np_dtype(src_dict[field_name])
        setattr(struct, field_name, value)
      else:             
        np_dtype = to_numpy_dtype(struct.fields[field_name].dtype)
        ndarray = np.array(src_dict[field_name], dtype=np_dtype)                    
        src_dict[field_name] = ndarray #prevent pointer of ndarray reassigned
        setattr(struct, field_name, ndarray.ctypes.data_as(field_type))    

def struct_to_dict(struct, order="F"):
  data = {}
  struct.fields
  for field_name in struct.fields.keys():
    np_dtype = to_numpy_dtype(struct.fields[field_name].dtype)  
    if struct.fields[field_name].shape == None:
      value = np_dtype(getattr(struct, field_name))
    else:
      ptr = getattr(struct, field_name)      
      shape = struct.fields[field_name].shape      
      value = make_nd_array(ptr, shape, dtype=np_dtype, order=order)    
    data[field_name] = value              
  return data

def make_nd_array(c_pointer, shape, dtype=np.float64, order='F', own_data=True):
  arr_size = np.prod(shape[:]) * np.dtype(dtype).itemsize 
  if sys.version_info.major >= 3:
    buf_from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    buf_from_mem.restype = ctypes.py_object
    buf_from_mem.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
    buffer = buf_from_mem(c_pointer, arr_size, 0x100)
  else:
    buf_from_mem = ctypes.pythonapi.PyBuffer_FromMemory
    buf_from_mem.restype = ctypes.py_object
    buffer = buf_from_mem(c_pointer, arr_size)
  arr = np.ndarray(tuple(shape[:]), dtype, buffer, order=order)
  if own_data and not arr.flags.owndata:
    return arr.copy().tolist()
  else:
      return arr.tolist()




class Task():
  def __init__(self, project_uuid, api, task_data=None):
    self.project_uuid = project_uuid
    self.api = api
    self.task_data = task_data
    self.output = {}

  def prepare(self):
    pass

  def run(self):
    pass

  def complete(self):
    pass

  def get_output(self):
    return self.output

  def terminate(self, exception_code):
    pass


import os 

def suppress():
    null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    save = os.dup(1), os.dup(2)
    os.dup2(null_fds[0], 1)
    os.dup2(null_fds[1], 2)
    return null_fds, save

def resume(null_fds, save):
	os.dup2(save[0], 1)
	os.dup2(save[1], 2)
	os.close(null_fds[0])
	os.close(null_fds[1])	    


import copy
import json

def merge(feature_collection, layer_label_prefix="Layer", algorithm="SAME_COORDINATES"):
  if algorithm == "SAME_COORDINATES":
    feature_dict = {}
    for feature in feature_collection["features"]:
      geometry_type = feature["geometry"]["type"]
      coordinates = feature["geometry"]["coordinates"]
      if geometry_type not in feature_dict:
        feature_dict[geometry_type] = {}

      coordinates_string = '#'.join(str(e) for e in coordinates)
      if coordinates_string not in feature_dict[geometry_type]:
        feature_dict[geometry_type][coordinates_string] = []      
      feature_dict[geometry_type][coordinates_string].append(feature)


    feature_collection["features"] = []
    for geometry_type in feature_dict:
      for coordinates_string in feature_dict[geometry_type]:
        features = feature_dict[geometry_type][coordinates_string]
        is_multi_layer = True if len(features) > 0 else False
        
        if is_multi_layer == True:
          align_feature = copy.deepcopy(features[0])
          align_feature["properties"] = {}
          for idx, feature  in enumerate(features):
            property_list = copy.deepcopy(list(feature["properties"].keys()))
           
            layer_label = "properties#{}-{}".format(layer_label_prefix, idx + 1)
            align_feature["properties"][layer_label] = {}
            for property in property_list:
              if idx == 0 and "ease-x#" in property:
                align_feature["properties"][property] = feature["properties"][property]
              elif "ease-x#" not in property:
                align_feature["properties"][layer_label][property] = feature["properties"][property]
                                  
          align_feature["properties"]["ease-x#multi-layer"] = True
 
        feature_collection["features"].append(align_feature)                  
      
  return json.dumps(feature_collection, ensure_ascii=False)



import requests
import json
import os
import pandas as pd

class EaseApi():
  def __init__(self, username=None, password=None, token=None, device_token=None):
    self.host = "https://app-backend.ease-x.com"
    self.token = token
    self.device_token = device_token
    self.username = username
    self.password = password
    self.s3_bucket_resource_uuid = None
    self.influx_bucket_resource_uuid = None
    if self.token == None and self.username != None and self.password != None:
      self._retrive_token()

  def _retrive_token(self):
    endpoint = f'{self.host}/api/commonauth/tokens'
    headers = {"content-type": "application/json"}
    data = {
      "username": self.username,
      "password": self.password,
      "method": "password"
    }
    r = requests.post(endpoint, json=data, headers=headers)
    self.token = r.json()["data"]["token"]

  def _get_project(self, project_uuid):
    endpoint = f'{self.host}/api/project/projects/{project_uuid}'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    r = requests.get(endpoint, headers=headers)
    data = r.json()["data"]
    if len(data["owner"]) > 0:
      return data["owner"][0]
    else:
      return data["member"][0]

  def _get_s3_bucket_resource_uuid(self, project_uuid):
    if self.s3_bucket_resource_uuid == None:
      project = self._get_project(project_uuid)
      for resource in project["resources"]:
        if resource["resource_type"] == "S3_BUCKET":
          return resource["uuid"]

  def _get_influx_bucket_resource_uuid(self, project_uuid):
    if self.influx_bucket_resource_uuid == None:
      project = self._get_project(project_uuid)
      for resource in project["resources"]:
        if resource["resource_type"] == "INFLUX_BUCKET":
          return resource["uuid"]

  def list_measurements(self, project_uuid):
    influx_bucket_resource_uuid = self._get_influx_bucket_resource_uuid(project_uuid)    
    endpoint = f'{self.host}/api/influx/measurements'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "parent_resource_uuid": influx_bucket_resource_uuid,
      "project_uuid": project_uuid
    }
    return requests.get(endpoint, params=params, headers=headers)


  def list_folders(self, project_uuid):
    s3_bucket_resource_uuid = self._get_s3_bucket_resource_uuid(project_uuid)    
    endpoint = f'{self.host}/api/s3/folders'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "parent_resource_uuid": s3_bucket_resource_uuid,
      "project_uuid": project_uuid
    }
    return requests.get(endpoint, params=params, headers=headers)

  def list_maplayers(self, project_uuid, mapset_resource_uuid): 
    endpoint = f'{self.host}/api/s3/maplayers'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "parent_resource_uuid": mapset_resource_uuid,
      "project_uuid": project_uuid
    }
    return requests.get(endpoint, params=params, headers=headers)

  def create_maplayer(self, project_uuid, mapset_resource_uuid, maplayer_name, settings=""):
    endpoint = f'{self.host}/api/s3/maplayers'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "parent_resource_uuid": mapset_resource_uuid,
      "maplayer_name": maplayer_name,
      "project_uuid": project_uuid,
      "settings": settings
    }
    return requests.post(endpoint, json=data, headers=headers)

  def create_staging_file(self, project_uuid, mapset_resource_uuid):
    endpoint = f'{self.host}/api/s3/stagingfiles'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "parent_resource_uuid": mapset_resource_uuid,
      "project_uuid": project_uuid,
    }
    return requests.post(endpoint, json=data, headers=headers)

  def delete_maplayer(self, project_uuid, maplayer_resource_uuid):
    endpoint = f'{self.host}/api/s3/maplayers'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "resource_uuid": maplayer_resource_uuid,
      "project_uuid": project_uuid,
    }
    return requests.delete(endpoint, json=data, headers=headers)

  def list_mapsets(self, project_uuid, folder_resource_uuid):
    endpoint = f'{self.host}/api/s3/mapsets'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "parent_resource_uuid": folder_resource_uuid,
      "project_uuid": project_uuid
    }
    return requests.get(endpoint, params=params, headers=headers)

  def create_mapset(self, project_uuid, folder_resource_uuid, mapset_name):
    endpoint = f'{self.host}/api/s3/mapsets'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "parent_resource_uuid": folder_resource_uuid,
      "mapset_name": mapset_name,
      "project_uuid": project_uuid,
    }
    return requests.post(endpoint, json=data, headers=headers)

  def delete_mapset(self, project_uuid, mapset_resource_uuid):
    endpoint = f'{self.host}/api/s3/mapsets'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "resource_uuid": mapset_resource_uuid,
      "project_uuid": project_uuid,
    }
    return requests.delete(endpoint, json=data, headers=headers)

  def create_folder(self, project_uuid, folder_name):
    bucket_resource_uuid = self._get_bucket_resource_uuid(project_uuid)
    endpoint = f'{self.host}/api/s3/folders'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "parent_resource_uuid": bucket_resource_uuid,
      "folder_name": folder_name,
      "project_uuid": project_uuid,
    }
    return requests.post(endpoint, json=data, headers=headers)

  def delete_folder(self, project_uuid, folder_resource_uuid):
    endpoint = f'{self.host}/api/s3/folders'
    headers = {"Authorization": f"Bearer {self.token}", "content-type": "application/json"}
    data = {
      "resource_uuid": folder_resource_uuid,
      "project_uuid": project_uuid,
    }
    return requests.delete(endpoint, json=data, headers=headers)

  def upload_maplayer_data(self, file_abs_location, filename, project_uuid, maplayer_resource_uuid):
    file_abs_path = "{}/{}".format(file_abs_location, filename)
    endpoint = f'{self.host}/api/s3/maplayer/data'
    files = [
      ('maplayer', (filename, open(file_abs_path,'rb'), 'application/octet-stream'))      
    ]
    data = {
      "project_uuid": project_uuid,
      "resource_uuid": maplayer_resource_uuid
    }    
    headers = {"Authorization": f"Bearer {self.token}"}
    return requests.put(endpoint, files=files, data=data, headers=headers)

  def upload_staging_file_data(self, file_abs_location, filename, project_uuid, maplayer_resource_uuid):
    file_abs_path = "{}/{}".format(file_abs_location, filename)
    endpoint = f'{self.host}/api/s3/stagingfile/data'
    files = [
      ('staging_file', (filename, open(file_abs_path,'rb'), 'application/octet-stream'))      
    ]
    data = {
      "project_uuid": project_uuid,
      "resource_uuid": maplayer_resource_uuid
    }    
    headers = {"Authorization": f"Bearer {self.token}"}
    return requests.put(endpoint, files=files, data=data, headers=headers)


  def get_maplayer(self, project_uuid, maplayer_resource_uuid):
    endpoint = f'{self.host}/api/s3/maplayers'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "resource_uuid": maplayer_resource_uuid,
      "project_uuid": project_uuid
    }
    return requests.get(endpoint, params=params, headers=headers)

  def fetch_data(self, project_uuid, maplayer_resource_uuid, is_json=False, output_filename=None):
    endpoint = f'{self.host}/api/s3/maplayer/data'
    headers = {"Authorization": f"Bearer {self.token}"}
    params = {
      "resource_uuid": maplayer_resource_uuid,
      "project_uuid": project_uuid,
    }
    if output_filename == None:
      output_filename = maplayer_resource_uuid
    with requests.get(endpoint, params=params, headers=headers, stream=True) as r:
      r.raise_for_status()
      with open(output_filename, 'wb+') as f:
        for chunk in r.iter_content(chunk_size=8192): 
          f.write(chunk)
    data = {}
    if is_json == True:
      f = open(output_filename, "r")
      data = json.load(f)
    return data

    
  def fetch_dataframe(self, project_uuid, measurement_resource_uuid, range=[], tags=[], fields=[], limit=100, keep=True):
    endpoint = f'{self.host}/api/influx/data'
    headers = {"Authorization": f"Bearer {self.token}"}
    data = {
      "resource_uuid": measurement_resource_uuid,
      "project_uuid": project_uuid,
      "start": range[0],
      "stop": range[1],
      "tags": tags,
      "fields": fields,
      "limit": limit
    }
    if os.path.isfile(measurement_resource_uuid) != True:
      r = requests.get(endpoint, json=data, headers=headers)
      if r.status_code == 200:     
        with open(measurement_resource_uuid, 'w') as f:        
          f.write(r.json()["data"])          
      else:
        raise Exception(r, "fetch error")

    with open(measurement_resource_uuid, 'r') as f:
      data = f.read()
      if data != "":
        data = pd.read_json(data, orient='index')

    if keep != True:
      os.remove(measurement_resource_uuid)
    return data

  def write_points(self, project_uuid, measurement_resource_uuid, points=[]):
    endpoint = f'{self.host}/api/influx/data'
    headers = {"Authorization": f"Bearer {self.device_token}", "content-type": "application/json"}
    data = {      
      "project_uuid": project_uuid,
      "resource_uuid": measurement_resource_uuid,
      "points": points
    }
    return requests.post(endpoint, json=data, headers=headers)